#ifndef __BSP_H__
#define __BSP_H__

#include <ubb/ubb.h>

/* Pin assignments (74HC4052):
 *
 * Sniffer  UBB   74HC4052 #1  74HC4052 #2  10K Resistor  M74HC595B1R  EM027AS012
 * -------  ----  -----------  -----------  ------------  -----------  ----------
 * DAT2     DAT2  13 (1Z)
 * CD       DAT3   3 (2Z)
 * CMD      CMD   10 (S0)      10 (S0)      IN       OUT  10 (SCLR#)
 * VCC      VDD   16 (VCC)     16 (VCC)                   16 (VCC)      2 (3V3)
 * CLK      CLK                                                         3 (SCK)
 * GND      GND    8 (GND)      8 (GND)                    8 (GND)      1 (GND)
 * DAT0     DAT0               13 (1Z)
 * DAT1     DAT1                3 (2Z)
 *
 * Multiplexer pin selection:
 *
 * {Z}Y{S1/S0}
 *
 * 1Z with S1=0, S0=0 => 1Y0 (output 1, channel 0)
 * 2Z with S1=0, S0=0 => 2Y0 (output 2, channel 0)
 *
 * 1Z with S1=0, S0=1 => 1Y1 (output 1, channel 1)
 * 2Z with S1=0, S0=1 => 2Y1 (output 2, channel 1)
 *
 * Connections to the shift register (channel 0):
 *
 * Sniffer  UBB   74HC4052 #1  M74HC595B1R
 * -------  ---   -----------  -----------
 * DAT2     DAT2  12 (1Y0)     12 (RCK)
 * CD       DAT3   1 (2Y0)     14 (SI)
 *
 *                74HC4052 #2
 *                -----------
 * DAT0     DAT0  12 (1Y0)     10 (SCLR#)
 * DAT1     DAT1   1 (2Y0)     11 (SCK)
 *
 * Connections to the display board (channel 1):
 *
 * Sniffer  UBB   74HC4052 #1               EM027AS012
 * -------  ---   -----------               ----------
 * DAT2     DAT2  14 (1Y1)                   4 (MOSI)
 * CD       DAT3   5 (2Y1)                   6 (SSEL)
 *
 *                74HC4052 #2
 *                -----------
 * DAT0     DAT0  14 (1Y1)                  11 (PWM)
 * DAT1     DAT1   5 (2Y1)                  12 (RESET)
 *
 * Connections from the shift register (providing persistent outputs for
 * channel 0):
 *
 *                             M74HC595B1R  EM027AS012
 *                             -----------  ----------
 *                             15 (QA)      13 (PANEL)
 *                              1 (QB)      14 (DIS)
 *
 * General connections:
 *
 * Sniffer  UBB   74HC4052     M74HC595B1R
 * -------  ----  --------     -----------
 * GND      GND    6 (E#)      13 (G#)
 *                 7 (VEE)
 *                 9 (S1)
 */

/* CMD/S0 = 0 */

#define SHIFT_SCLR  UBB_DAT0
#define SHIFT_SCK   UBB_DAT1
#define SHIFT_RCK   UBB_DAT2
#define SHIFT_SI    UBB_DAT3

/* CMD/S1 = 1 */

#define EPD_PWM     UBB_DAT0
#define EPD_RESET   UBB_DAT1
#define EPD_MOSI    UBB_DAT2
#define EPD_SSEL    UBB_DAT3

/* Channel-insensitive. */

#define MUX_S0      UBB_CMD
#define EPD_SCK     UBB_CLK

/* NanoNote-specific functions. */

void shift_select();
void shift_update();
void shift_init();
void shift_reinit();
void shift_in(uint8_t v);
void epd_select();

/* Board-specific interface functions. */

void bsp_init();
void bsp_InitDisplayHardware();
void bsp_PowerOn();
void bsp_pwm_active(uint16_t delayInMs);
void bsp_Shutdown();
void bsp_spiInit();
void bsp_writeToDisplay(uint8_t *data, uint16_t len);
void bsp_delayMs(uint32_t ms);
void bsp_delayUs(uint32_t us);
uint32_t bsp_getMsTicks();
int16_t bsp_getTemp();

#endif /* __BSP_H__ */
